unit JPEG_IO;

{
Revision history:

2000 Nov 03  First version
2000 Nov 20  Add more error checking on read
2000 Dec 04  Remove comment field until Intel library fixed
2001 Apr 08  Add ability to read scale down images for thumbnails


Copyright  David J Taylor, Edinburgh
Web site:  www.satsignal.net
E-mail:    davidtaylor@writeme.com
}

interface

uses Windows, SysUtils, Classes, Graphics, IJL;   // Intel JPEG Library

type
  TIJLScale = (ijlFullSize, ijlHalf, ijlQuarter, ijlEighth);


function SaveBmpToJpegFile (const bitmap: TBitmap;  const filename: string;
                    const quality: integer;  const comment: string): boolean;

function LoadBmpFromJpegFile (bitmap: TBitmap;  const filename: string;
                              const set_pixel_format: boolean;
                              const scale: TIJLScale = ijlFullSize): boolean;


implementation

function SaveBmpToJpegFile (const bitmap: TBitmap;  const filename: string;
                    const quality: integer;  const comment: string): boolean;
var
  jcprops : TJPEG_CORE_PROPERTIES;
  iWidth, iHeight, iNChannels : Integer;
  iStatus: integer;
  DIB: TDIBSection;
begin
  // Should raise an exception if it fails
  Result := False;

  // Initialise the JPEG library
  FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
  iStatus := ijlInit (@jcprops);
  if iStatus = IJL_OK then
  try
    jcprops.jprops.jpeg_comment := PChar (comment);
    jcprops.jprops.jpeg_comment_size := Length (comment);
    FillChar (DIB, SizeOf(DIB), 0);
    GetObject (Bitmap.Handle, SizeOf (DIB), @DIB);
    iWidth  := DIB.dsBm.bmWidth;
    iHeight := DIB.dsBm.bmHeight;
    case bitmap.PixelFormat of
      pf8bit: iNChannels := 1;
      pf24bit: iNChannels := 3;
    else
      Raise EInvalidOperation.Create ('Cannot save bitmap as JPEG with specified PixelFormat');
    end;

    jcprops.DIBWidth := iWidth;
    jcprops.DIBHeight := -iHeight;
    jcprops.DIBChannels := iNChannels;
    case bitmap.PixelFormat of
       pf8bit: jcprops.DIBColor := IJL_G;
      pf24bit: jcprops.DIBColor := IJL_BGR;
    end;
    jcprops.DIBPadBytes := ((((iWidth*iNChannels)+3) div 4)*4)-(iWidth*iNChannels);
    jcprops.DIBBytes := PByte (DIB.dsBm.bmBits);

    jcprops.JPGFile := PChar (FileName);
    jcprops.JPGWidth := iWidth;
    jcprops.JPGHeight := iHeight;
    jcprops.JPGChannels := 3;
    jcprops.JPGColor := IJL_YCBCR;
    jcprops.jquality := quality;
    Result := IJL_OK = ijlWrite (@jcprops, IJL_JFILE_WRITEWHOLEIMAGE);
    ijlFree (@jcprops);
  except
  end;
  if not Result then Raise (EWriteError.Create ('Error writing JPEG to file'));
end;


function LoadBmpFromJpegFile (bitmap: TBitmap;  const filename: string;
                              const set_pixel_format: boolean;
                              const scale: TIJLScale = ijlFullSize): boolean;

var
  iWidth, iHeight, iNChannels: Integer;
  iDIBChannels: integer;
  iStatus: integer;
  jcprops: TJPEG_CORE_PROPERTIES;
  DIB: TDIBSection;
begin
  Result := False;
  if not Assigned (bitmap) then Exit;
  if not FileExists (filename) then Exit;

  case bitmap.PixelFormat of
    pf8Bit: iDIBchannels := 1;
    pf24Bit: iDIBchannels := 3;
  else
    iDIBchannels := 0;
  end;
  if (iDIBChannels = 0) and (not set_pixel_format) then Exit;

  // Initialise the JPEG library
  FillChar (jcprops, SizeOf (jcprops), 0);  // Just to be sure...
  iStatus := ijlInit (@jcprops);
  if iStatus = IJL_OK then
  try
    jcprops.JPGFile := PChar (FileName);
    iStatus := ijlRead (@jcprops, IJL_JFILE_READPARAMS);
    if iStatus = IJL_OK then
      begin
      iWidth := jcprops.JPGWidth;
      iHeight := jcprops.JPGHeight;
      case scale of
        ijlHalf: begin
                 iWidth := (iWidth + 1) div 2;
                 iHeight := (iHeight + 1) div 2;
                 end;
        ijlQuarter: begin
                 iWidth := (iWidth + 3) div 4;
                 iHeight := (iHeight + 3) div 4;
                 end;
        ijlEighth: begin
                 iWidth := (iWidth + 7) div 8;
                 iHeight := (iHeight + 7) div 8;
                 end;
      end;
      iNChannels := jcprops.JPGChannels;
      if (iNChannels = 1) or (iNChannels = 3) then
        begin
        bitmap.Height := 0;
        bitmap.Width := 0;
        if set_pixel_format then
          begin
          iDIBChannels := iNChannels;
          if iDIBChannels = 1
            then bitmap.PixelFormat := pf8Bit
            else bitmap.PixelFormat := pf24Bit;
          end;
        bitmap.Width := iWidth;
        bitmap.Height := iHeight;
        FillChar (DIB, SizeOf (DIB), 0);
        iStatus := GetObject (bitmap.Handle, SizeOf (DIB), @DIB);
        if iStatus <> 0 then
          begin
          jcprops.DIBWidth := iWidth;
          jcprops.DIBHeight := -iHeight;
          jcprops.DIBChannels := iDIBChannels;
          if iDIBchannels = 3
            then jcprops.DIBColor := IJL_BGR
            else jcprops.DIBColor := IJL_G;
          jcprops.DIBPadBytes := ((((iWidth*iDIBChannels)+3) div 4)*4)-(iWidth*iDIBChannels);
          jcprops.DIBBytes := PByte (DIB.dsBm.bmBits);

          case scale of
            ijlHalf: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEHALF);
            ijlQuarter: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEQUARTER);
            ijlEighth: iStatus := ijlRead (@jcprops, IJL_JFILE_READONEEIGHTH);
          else
            iStatus := ijlRead (@jcprops, IJL_JFILE_READWHOLEIMAGE);
          end;

          if iStatus >= 0 then
            begin
            bitmap.Modified := True;
            Result := True;
            end;
          end;
        end;
      end;
  finally
    ijlFree (@jcprops);
  end;
end;

end.

