//  Statistics Library


UNIT StatisticsLibrary;

INTERFACE

  USES
    Windows,  // TRGBTriple
    Graphics, // TBitmap
    IEEE754,  // NaN
    Math;     // MaxDouble

  TYPE

   // ===================================================================

   TDescriptiveStatistics =
   CLASS(TObject)
     PROTECTED
       FCount       :  INTEGER;

       FMaxValue    :  DOUBLE;
       FMinValue    :  DOUBLE;

       FSum         :  DOUBLE;
       FSumOfSquares:  DOUBLE;

     PUBLIC

       PROPERTY  Count   :  INTEGER   READ FCount;
       PROPERTY  MaxValue:  DOUBLE    READ FMaxValue;
       PROPERTY  MinValue:  DOUBLE    READ FMinValue;
       PROPERTY  Sum     :  DOUBLE    READ FSum;

       CONSTRUCTOR Create;

       PROCEDURE NextValue(CONST x:  DOUBLE);

       FUNCTION MeanValue        :  DOUBLE;
       FUNCTION StandardDeviation:  DOUBLE;

       PROCEDURE ResetValues;

   END;


  // ===================================================================

   // Formulas from HP-45 Applications Book, pp. 72, 77, 1974
   TLinearRegression =
   CLASS(TObject)
     PROTECTED
       FCount      :  INTEGER;

       FSumX       :  DOUBLE;
       FSumXSquared:  DOUBLE;

       FSumY       :  DOUBLE;
       FSumYSquared:  DOUBLE;

       FSumXY      :  DOUBLE;

     PUBLIC
       CONSTRUCTOR Create;
       FUNCTION  CorrelationCoefficient:  DOUBLE;
       PROCEDURE GetLineCoefficients(VAR A,B:  DOUBLE);
       FUNCTION  RSquared:  DOUBLE;
       PROCEDURE NextValue(CONST x, y:  DOUBLE);
       PROCEDURE ResetValues;

   END;


   // ===================================================================

   FUNCTION MedianInteger(x:  ARRAY OF INTEGER):  INTEGER;

   PROCEDURE CorrelateBitmaps(CONST Bitmap1, Bitmap2:  TBitmap;
                              VAR RedCorrelation,
                                  GreenCorrelation,
                                  BlueCorrelation,
                                  IntensityCorrelation:  DOUBLE);


IMPLEMENTATION

  USES
    SysUtils;   // EInvalidOp


// ==== TDescriptiveStatistics =========================================

  CONSTRUCTOR TDescriptiveStatistics.Create;
  BEGIN
    ResetValues
  END {Create};


  PROCEDURE TDescriptiveStatistics.NextValue (CONST x:  DOUBLE);
  BEGIN
    INC(FCount);
    FSum := FSum + x;
    FSumOfSquares := FSumOfSquares + x*x;

    IF   x > FMaxValue
    THEN FmaxValue := x;

    IF  x < FMinValue
    THEN FMinValue := x
  END {NextNumber};


  FUNCTION TDescriptiveStatistics.MeanValue:  DOUBLE;
  BEGIN
    IF   FCount = 0
    THEN RESULT := NAN
    ELSE RESULT := FSum / FCount
  END {MeanValue};


  FUNCTION TDescriptiveStatistics.StandardDeviation:  DOUBLE;
  BEGIN
    IF   FCount < 2
    THEN RESULT := NAN
    ELSE RESULT := SQRT( (FSumOfSquares - (FSum*FSum / FCount)) /
                         (FCount - 1) )
  END {StandardDeviation};


  PROCEDURE TDescriptiveStatistics.ResetValues;
  BEGIN
    FCount := 0;

    FMinValue :=  MaxDouble;
    FMaxValue := -MaxDouble;

    FSum          := 0.0;
    FSumOfSquares := 0.0
  END {ResetValues};

// ==== TLinearRegression ==============================================

  CONSTRUCTOR TLinearRegression.Create;
  BEGIN
    ResetValues
  END {Create};


  {Formulas from HP-45 Applications Book, p. 77, 1974}
  PROCEDURE TLinearRegression.GetLineCoefficients(VAR A,B:  DOUBLE);
  BEGIN
    IF   FCount < 1
    THEN BEGIN
      A := NAN;
      B := NAN
    END
    ELSE BEGIN
       A := (FSumXY - FSumX*FSumY/FCount) /
            (FSumXSquared - FSumX*FSumX/FCount);
       B := (FSumY - A*FSumX)/FCount
    END
  END {GetLineCoefficients};


  // Formula from HP-45 Applications Book, p. 72, 1974
  FUNCTION  TLinearRegression.CorrelationCoefficient:  DOUBLE;
    VAR
      numerator  :  DOUBLE;
      denominator:  DOUBLE;
  BEGIN
    IF   FCount < 2
    THEN RESULT := NAN
    ELSE BEGIN
      numerator   := FSumXY - FSumX*FSumY/FCount;
      denominator := SQRT( (FSumXSquared - FSumX*FSumX/FCount)*
                         (FSumYSquared - FSumY*FSumY/FCount) );
      TRY
        RESULT :=  numerator / denominator
      EXCEPT
        On EInvalidOp
          DO  IF   (numerator = 0.0) AND (denominator = 0.0)
              THEN RESULT := 1.0    // all zeros have "perfect" correlation
              ELSE RESULT := NAN;
      END
    END
  END {CorrelationCoefficient};


  FUNCTION  TLinearRegression.RSquared:  DOUBLE;
  BEGIN
    RESULT := 0.0    {FIX THIS}
  END {RSquared};


  PROCEDURE TLinearRegression.NextValue(CONST x, y:  DOUBLE);
  BEGIN
    INC(FCount);

    FSumX        := FSumX + x;
    FSumXSquared := FSumXSquared + x*x;

    FSumY        := FSumY + y;
    FSumYSquared := FSumYSquared + y*y;

    FSumXY       := FSumXY + x*y
  END {NextValue};


  PROCEDURE TLinearRegression.ResetValues;
  BEGIN
    FCount       := 0;

    FSumX        := 0.0;
    FSumXSquared := 0.0;

    FSumY        := 0.0;
    FSumYSquared := 0.0;

    FSumXY       := 0.0
  END {ResetValues};

// ==== Median =========================================================

  FUNCTION MedianInteger (x:  ARRAY OF INTEGER):  INTEGER;
    VAR
      i        :  INTEGER;
      j        :  INTEGER;
      Middle   :  INTEGER;
      Temporary:  INTEGER;
  BEGIN
    {Use truncated selection sort to find median}

    Middle := (High(x)+1) DIV 2;

    FOR i := 0 TO Middle DO
    BEGIN
      FOR j := 1 TO High(x)-i DO
      BEGIN
        IF   x[j] > x[j-1]
        THEN BEGIN
          Temporary := x[j];
          x[j] := x[j-1];
          x[j-1] := Temporary
        END
      END

    END;

    IF   Odd(High(x))
    THEN BEGIN
      {When High(x) is Odd, there are an even number of elements in array.
       Define median as average of two middle values.}
       RESULT := (x[middle] + x[middle-1]) DIV 2
    END
    ELSE BEGIN
       {When High(x) is Even, there are an odd number of elements in array.
       Median is the middle value.}
       RESULT := x[middle]
    END
  END {MedianInteger};


// =====================================================================

  PROCEDURE CorrelateBitmaps(CONST Bitmap1, Bitmap2:  TBitmap;
                             VAR RedCorrelation,
                                 GreenCorrelation,
                                 BlueCorrelation,
                                 IntensityCorrelation:  DOUBLE);
  CONST
    PixelMaxCount = 65536;

  TYPE
    TRGBArray = ARRAY[0..PixelMaxCount-1] OF TRGBTriple;
    pRGBArray = ^TRGBArray;

    VAR
      CorrelationBlue        :  TLinearRegression;
      CorrelationGreen       :  TLinearRegression;
      CorrelationRed         :  TLinearRegression;
      CorrelationIntensity   :  TLinearRegression;
      i                      :  INTEGER;
      j                      :  INTEGER;
      row1                   :  pRGBArray;
      row2                   :  pRGBArray;
  BEGIN
    IF  (Bitmap1.Width       <> Bitmap2.Width)       OR
        (Bitmap1.Height      <> Bitmap2.Height)      OR
        (Bitmap1.PixelFormat <> Bitmap2.PixelFormat) OR
        (Bitmap1.PixelFormat <> pf24bit)
    THEN BEGIN
       RedCorrelation       := NAN;
       GreenCorrelation     := NAN;
       BlueCorrelation      := NAN;
       IntensityCorrelation := NAN
    END
    ELSE BEGIN
      CorrelationRed       := TLinearRegression.Create;
      CorrelationGreen     := TLinearRegression.Create;
      CorrelationBlue      := TLinearRegression.Create;
      CorrelationIntensity := TLinearRegression.Create;
      TRY
        FOR j := 0 TO Bitmap1.Height-1 DO
        BEGIN
          row1       := Bitmap1.ScanLine[j];
          row2       := Bitmap2.ScanLine[j];

          FOR i := 0 TO Bitmap1.Width-1 DO
          BEGIN
            CorrelationRed.NextValue(row1[i].rgbtRed,  row2[i].rgbtRed);
            CorrelationGreen.NextValue(row1[i].rgbtGreen,  row2[i].rgbtGreen);
            CorrelationBlue.NextValue(row1[i].rgbtBlue,  row2[i].rgbtBlue);

            CorrelationIntensity.NextValue(
                  (row1[i].rgbtRed + row1[i].rgbtGreen + row1[i].rgbtBlue) / 3,
                  (row2[i].rgbtRed + row2[i].rgbtGreen + row2[i].rgbtBlue) / 3)
          END
        END;

        RedCorrelation       := CorrelationRed.CorrelationCoefficient;
        GreenCorrelation     := CorrelationRed.CorrelationCoefficient;
        BlueCorrelation      := CorrelationRed.CorrelationCoefficient;
        IntensityCorrelation := CorrelationRed.CorrelationCoefficient

      FINALLY
        CorrelationRed.Free;
        CorrelationGreen.Free;
        CorrelationBlue.Free;
        CorrelationIntensity.Free
      END

  END
END {CompareBitmaps};


// =====================================================================

END.
